# Copyright 2022 DSE lab.  All rights reserved.

import numpy as np
from typing import *
import torch
import torch.nn as nn
import torch.nn.functional as F
import math

def kld_loss(p, q):
    def kld(target, pred):
        return torch.mean(torch.sum(target*torch.log(target/(pred+1e-6)), dim=1))
    loss = kld(p, q)
    return loss

###############################################
##    Created by BABEL
###############################################
class BCELoss(nn.BCELoss):
    """Custom BCE loss that can correctly ignore the encoded latent space output"""

    def forward(self, x, target):
        import ipdb
        ipdb.set_trace()
        input = x[0]
        return F.binary_cross_entropy(
            input, target, weight=self.weight, reduction=self.reduction
        )

class MSELoss(nn.MSELoss):
    """Custom MSE loss that can correctly ignore the encoded latent space output"""

    def forward(self, x: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        return F.mse_loss(x[0], target, reduction=self.reduction)

class RMSELoss(nn.MSELoss):
    """Custom RMSE loss that can correctly ignore the encoded latent space output"""

    def forward(self, x: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        return torch.sqrt(F.mse_loss(x[0], target, reduction=self.reduction))

class DistanceProbLoss(nn.Module):
    """
    Analog of above log prob loss, but using distances
    May be useful for aligning latent spaces
    """

    def __init__(self, weight: float = 5.0, norm: int = 1):
        super(DistanceProbLoss, self).__init__()
        assert weight > 0
        self.weight = weight
        self.norm = norm

    def forward(self, x, target_z):
        z, logp = x[:2]
        d = F.pairwise_distance(
            z,
            target_z,
            p=self.norm,
            eps=1e-6,
            keepdim=False,  # Default value
        )
        if len(d.shape) == 2:
            d = torch.mean(d, dim=1)  # Drop 1 dimension
        per_ex = self.weight * d - logp
        retval = torch.mean(per_ex)
        if retval != retval:
            raise ValueError("NaN")
        return retval
        return torch.mean(d)

class NegativeBinomialLoss(nn.Module):
    """
    Negative binomial loss. Preds should be a tuple of (mean, dispersion)
    """

    def __init__(
        self,
        scale_factor: float = 1.0,
        eps: float = 1e-10,
        l1_lambda: float = 0.0,
        mean: bool = True,
    ):
        super(NegativeBinomialLoss, self).__init__()
        self.loss = negative_binom_loss(
            scale_factor=scale_factor,
            eps=eps,
            mean=mean,
            debug=True,
        )
        self.l1_lambda = l1_lambda

    def forward(self, preds, target):
        preds, theta = preds[:2]
        l = self.loss(
            preds=preds,
            theta=theta,
            truth=target,
        )
        encoded = preds[:-1]
        l += self.l1_lambda * torch.abs(encoded).sum()
        return l


class ZeroInflatedNegativeBinomialLoss(nn.Module):
    """
    ZINB loss. Preds should be a tuple of (mean, dispersion, dropout)
    General notes:
    total variation seems to do poorly (at least for atacseq)
    """

    def __init__(
        self,
        ridge_lambda: float = 0.0,
        tv_lambda: float = 0.0,
        l1_lambda: float = 0.0,
        eps: float = 1e-10,
        scale_factor: float = 1.0,
        debug: bool = True,
    ):
        super(ZeroInflatedNegativeBinomialLoss, self).__init__()
        self.loss = zero_inflated_negative_binom_loss(
            ridge_lambda=ridge_lambda,
            tv_lambda=tv_lambda,
            eps=eps,
            scale_factor=scale_factor,
            debug=debug,
        )
        self.l1_lambda = l1_lambda

    def forward(self, preds, target):
        preds, theta, pi = preds[:3]
        l = self.loss(
            preds=preds,
            theta_disp=theta,
            pi_dropout=pi,
            truth=target,
        )
        encoded = preds[:-1]
        l += self.l1_lambda * torch.abs(encoded).sum()
        return l

class PairedLoss(nn.Module):
    """
    Paired loss function. Automatically unpacks and encourages the encoded representation to be similar
    using a given distance function. link_strength parameter controls how strongly we encourage this
    loss2_weight controls how strongly we weight the second loss, relative to the first
    A value of 1.0 indicates that they receive equal weight, and a value larger indicates
    that the second loss receives greater weight.
    link_func should be a callable that takes in the two encoded representations and outputs a metric
    where a larger value indicates greater divergence
    """

    def __init__(
        self,
        loss1=NegativeBinomialLoss,
        loss2=ZeroInflatedNegativeBinomialLoss,
        link_func=lambda x, y: (x - y).abs().mean(),
        link_strength=1e-3,
    ):
        super(PairedLoss, self).__init__()
        self.loss1 = loss1()
        self.loss2 = loss2()
        self.link = link_strength
        self.link_f = link_func

        self.warmup = SigmoidWarmup(
            midpoint=1000,
            maximum=link_strength,
        )

    def forward(self, preds, target):
        """Unpack and feed to each loss, averaging at end"""
        preds1, preds2 = preds
        target1, target2 = target

        loss1 = self.loss1(preds1, target1)
        loss2 = self.loss2(preds2, target2)
        retval = loss1 + loss2

        # Align the encoded representation assuming the last output is encoded representation
        encoded1 = preds1[-1]
        encoded2 = preds2[-1]
        if self.link > 0:
            l = next(self.warmup)
            if l > 1e-6:
                d = self.link_f(encoded1, encoded2).mean()
                retval += l * d

        return retval


class PairedLossInvertible(nn.Module):
    """
    Paired loss function with additional invertible (RealNVP) layer loss
    Loss 1 is for the first autoencoder
    Loss 2 is for the second autoencoder
    Loss 3 is for the invertible network at bottleneck
    """

    def __init__(
        self,
        loss1=NegativeBinomialLoss,
        loss2=ZeroInflatedNegativeBinomialLoss,
        loss3=DistanceProbLoss,
        link_func=lambda x, y: (x - y).abs().mean(),
        link_strength=1e-3,
        inv_strength=1.0,
    ):
        super(PairedLossInvertible, self).__init__()
        self.loss1 = loss1()
        self.loss2 = loss2()
        self.loss3 = loss3()
        self.link = link_strength
        self.link_f = link_func

        # self.link_warmup = layers.SigmoidWarmup(
        #     midpoint=1000,
        #     maximum=link_strength,
        # )
        self.link_warmup = DelayedLinearWarmup(
            delay=1000,
            inc=5e-3,
            t_max=link_strength,
        )

        self.inv_warmup = DelayedLinearWarmup(
            delay=2000,
            inc=5e-3,
            t_max=inv_strength,
        )

    def forward(self, preds, target):
        """Unpack and feed to each loss"""
        # Both enc1_pred and enc2_pred are tuples of 2 values
        preds1, preds2, (enc1_pred, enc2_pred) = preds
        target1, target2 = target

        loss1 = self.loss1(preds1, target1)
        loss2 = self.loss2(preds2, target2)
        retval = loss1 + loss2

        # Align the encoded representations
        encoded1 = preds1[-1]
        encoded2 = preds2[-1]
        if self.link > 0:
            l = next(self.link_warmup)
            if l > 1e-6:
                d = self.link_f(encoded1, encoded2).mean()
                retval += l * d

        # Add a term for invertible network
        inv_loss1 = self.loss3(enc1_pred, enc2_pred[0])
        inv_loss2 = self.loss3(enc2_pred, enc1_pred[0])
        retval += next(self.inv_warmup) * (inv_loss1 + inv_loss2)

        return retval


class QuadLoss(PairedLoss):
    """
    Paired loss, but for the spliced autoencoder with 4 outputs
    """

    def __init__(
        self,
        loss1=NegativeBinomialLoss,
        loss2=BCELoss,
        loss2_weight: float = 3.0,
        cross_weight: float = 1.0,
        cross_warmup_delay: int = 0,
        link_strength: float = 0.0,
        link_func: Callable = lambda x, y: (x - y).abs().mean(),
        link_warmup_delay: int = 0,
        record_history: bool = False,
    ):
        super(QuadLoss, self).__init__()
        self.loss1 = loss1()
        self.loss2 = loss2()
        self.loss2_weight = loss2_weight
        self.history = []  # Eventually contains list of tuples per call
        self.record_history = record_history

        if link_warmup_delay:
            self.warmup = SigmoidWarmup(
                midpoint=link_warmup_delay,
                maximum=link_strength,
            )
            # self.warmup = layers.DelayedLinearWarmup(
            #     delay=warmup_delay,
            #     t_max=link_strength,
            #     inc=1e-3,
            # )
        else:
            self.warmup = NullWarmup(t_max=link_strength)
        if cross_warmup_delay:
            self.cross_warmup = SigmoidWarmup(
                midpoint=cross_warmup_delay,
                maximum=cross_weight,
            )
        else:
            self.cross_warmup = NullWarmup(t_max=cross_weight)

        self.link_strength = link_strength
        self.link_func = link_func

    def get_component_losses(self, preds, target):
        """
        Return the four losses that go into the overall loss, without scaling
        """
        preds11, preds12, preds21, preds22 = preds
        if not isinstance(target, (list, tuple)):
            # Try to unpack into the correct parts
            target = torch.split(
                target, [preds11[0].shape[-1], preds22[0].shape[-1]], dim=-1
            )
        target1, target2 = target  # Both are torch tensors

        loss11 = self.loss1(preds11, target1)
        loss21 = self.loss1(preds21, target1)
        loss12 = self.loss2(preds12, target2)
        loss22 = self.loss2(preds22, target2)

        return loss11, loss21, loss12, loss22

    def forward(self, preds, target):
        loss11, loss21, loss12, loss22 = self.get_component_losses(preds, target)
        if self.record_history:
            detensor = lambda x: x.detach().cpu().numpy().item()
            self.history.append([detensor(l) for l in (loss11, loss21, loss12, loss22)])

        loss = loss11 + self.loss2_weight * loss22
        loss += next(self.cross_warmup) * (loss21 + self.loss2_weight * loss12)

        if self.link_strength > 0:
            l = next(self.warmup)
            if l > 1e-6:  # If too small we disregard
                preds11, preds12, preds21, preds22 = preds
                encoded1 = preds11[-1]  # Could be preds12
                encoded2 = preds22[-1]  # Could be preds21
                d = self.link_func(encoded1, encoded2)
                loss += self.link_strength * d
        return loss


def scvi_log_nb_positive(x, mu, theta, eps=1e-8):
    """
    Taken from scVI log_likelihood.py - scVI invocation is:
    reconst_loss = -log_nb_positive(x, px_rate, px_r).sum(dim=-1)
    scVI decoder outputs px_scale, px_r, px_rate, px_dropout
    px_scale is subject to Softmax
    px_r is just a Linear layer
    px_rate = torch.exp(library) * px_scale
    mu = mean of NB
    theta = indverse dispersion parameter
    Here, x appears to correspond to y_true in the below negative_binom_loss (aka the observed counts)
    """
    # if theta.ndimension() == 1:
    #     theta = theta.view(
    #         1, theta.size(0)
    #     )  # In this case, we reshape theta for broadcasting

    log_theta_mu_eps = torch.log(theta + mu + eps)
    res = (
        theta * (torch.log(theta + eps) - log_theta_mu_eps)
        + x * (torch.log(mu + eps) - log_theta_mu_eps)
        + torch.lgamma(x + theta)
        - torch.lgamma(theta)  # Present (in negative) for DCA
        - torch.lgamma(x + 1)
    )

    return res.mean()


def scvi_log_zinb_positive(x, mu, theta, pi, eps=1e-8):
    """
    https://github.com/YosefLab/scVI/blob/6c9f43e3332e728831b174c1c1f0c9127b77cba0/scvi/models/log_likelihood.py#L206
    """
    # theta is the dispersion rate. If .ndimension() == 1, it is shared for all cells (regardless of batch or labels)
    if theta.ndimension() == 1:
        theta = theta.view(
            1, theta.size(0)
        )  # In this case, we reshape theta for broadcasting

    softplus_pi = F.softplus(-pi)
    log_theta_eps = torch.log(theta + eps)
    log_theta_mu_eps = torch.log(theta + mu + eps)
    pi_theta_log = -pi + theta * (log_theta_eps - log_theta_mu_eps)

    case_zero = F.softplus(pi_theta_log) - softplus_pi
    mul_case_zero = torch.mul((x < eps).type(torch.float32), case_zero)

    case_non_zero = (
        -softplus_pi
        + pi_theta_log
        + x * (torch.log(mu + eps) - log_theta_mu_eps)  # Found above
        + torch.lgamma(x + theta)  # Found above
        - torch.lgamma(theta)  # Found above
        - torch.lgamma(x + 1)  # Found above
    )
    mul_case_non_zero = torch.mul((x > eps).type(torch.float32), case_non_zero)

    res = mul_case_zero + mul_case_non_zero

    return res.mean()


def negative_binom_loss(
    scale_factor: float = 1.0,
    eps: float = 1e-10,
    mean: bool = True,
    debug: bool = False,
) -> Callable:
    """
    Return a function that calculates the binomial loss
    https://github.com/theislab/dca/blob/master/dca/loss.py
    combination of the Poisson distribution and a gamma distribution is a negative binomial distribution
    """

    def loss(preds, theta, truth):
        """Calculates negative binomial loss as defined in the NB class in link above"""
        y_true = truth
        y_pred = preds * scale_factor

        if debug:  # Sanity check before loss calculation
            assert not torch.isnan(y_pred).any(), y_pred
            assert not torch.isinf(y_pred).any(), y_pred
            assert not (y_pred < 0).any()  # should be non-negative
            assert not (theta < 0).any()

        # Clip theta values
        theta = torch.clamp(theta, max=1e6)

        t1 = (
            torch.lgamma(theta + eps)
            + torch.lgamma(y_true + 1.0)
            - torch.lgamma(y_true + theta + eps)
        )
        t2 = (theta + y_true) * torch.log1p(y_pred / (theta + eps)) + (
            y_true * (torch.log(theta + eps) - torch.log(y_pred + eps))
        )
        if debug:  # Sanity check after calculating loss
            assert not torch.isnan(t1).any(), t1
            assert not torch.isinf(t1).any(), (t1, torch.sum(torch.isinf(t1)))
            assert not torch.isnan(t2).any(), t2
            assert not torch.isinf(t2).any(), t2

        retval = t1 + t2
        if debug:
            assert not torch.isnan(retval).any(), retval
            assert not torch.isinf(retval).any(), retval

        return torch.mean(retval) if mean else retval

    return loss


def zero_inflated_negative_binom_loss(
    ridge_lambda: float = 0.0,
    tv_lambda: float = 0.0,
    eps: float = 1e-10,
    scale_factor: float = 1.0,
    debug: bool = False,
):
    """
    Return a function that calculates ZINB loss
    https://github.com/theislab/dca/blob/master/dca/loss.py
    """
    nb_loss_func = negative_binom_loss(
        mean=False, eps=eps, scale_factor=scale_factor, debug=debug
    )

    def loss(preds, theta_disp, pi_dropout, truth):
        if debug:
            assert not (pi_dropout > 1.0).any()
            assert not (pi_dropout < 0.0).any()
        nb_case = nb_loss_func(preds, theta_disp, truth) - torch.log(
            1.0 - pi_dropout + eps
        )

        y_true = truth
        y_pred = preds * scale_factor
        theta = torch.clamp(theta_disp, max=1e6)

        zero_nb = torch.pow(theta / (theta + y_pred + eps), theta)
        zero_case = -torch.log(pi_dropout + ((1.0 - pi_dropout) * zero_nb) + eps)
        result = torch.where(y_true < 1e-8, zero_case, nb_case)

        # Ridge regularization on pi dropout term
        ridge = ridge_lambda * torch.pow(pi_dropout, 2)
        result += ridge

        # Total variation regularization on pi dropout term
        tv = tv_lambda * total_variation(pi_dropout)
        result += tv

        retval = torch.mean(result)
        # if debug:
        #     assert retval.item() > 0
        return retval

    return loss

def total_variation(x):
    """
    Given a 2D input (where one dimension is a batch dimension, the actual values are
    one dimensional) compute the total variation (within a 1 position shift)
    """
    t = torch.sum(torch.abs(x[:, :-1] - x[:, 1:]))
    return t


class Warmup(object):  # Doesn't have to be nn.Module because it's not learned
    """
    Warmup layer similar to
    Sonderby 2016 - Linear deterministic warm-up
    """

    def __init__(self, inc: float = 5e-3, t_max: float = 1.0):
        self.t = 0.0
        self.t_max = t_max
        self.inc = inc
        self.counter = 0  # Track number of times called next

    def __iter__(self):
        return self

    def __next__(self):
        retval = self.t
        t_next = self.t + self.inc
        self.t = min(t_next, self.t_max)
        self.counter += 1
        return retval

class DelayedLinearWarmup(object):
    """
    """

    def __init__(self, delay: int = 2000, inc: float = 5e-3, t_max: float = 1.0):
        self.t = 0.0
        self.t_max = t_max
        self.inc = inc
        self.delay = delay
        self.counter = 0

    def __iter__(self):
        return self

    def __next__(self):
        self.counter += 1
        retval = self.t
        if self.counter < self.delay:
            return retval
        self.t = min(self.t + self.inc, self.t_max)
        return retval

class SigmoidWarmup(object):
    """
    Sigmoid warmup
    Midpoints defines the number of iterations before we hit 0.5
    Scale determines how quickly we hit 1 after this point
    """

    def __init__(self, midpoint: int = 500, scale: float = 0.1, maximum: float = 1.0):
        self.midpoint = midpoint
        self.scale = scale
        self.maximum = maximum
        self.counter = 0
        self.t = 0.0

    def __iter__(self):
        return self

    def __next__(self):
        retval = self.t
        t_next = 1.0 / (1.0 + np.exp(-self.scale * (self.counter - self.midpoint)))
        self.t = t_next
        self.counter += 1
        return self.maximum * retval

class NullWarmup(Warmup):
    """
    No warmup - but provides a consistent API
    """

    def __init__(self, delay: int = 0, t_max: float = 1.0):
        self.val = t_max

    def __next__(self):
        return self.val

####################################
## Created by scMVAE
####################################


def GMM_loss(gamma, c_params, z_params):
    """
    L elbo(x) = Eq(z,c|x)[ log p(x|z) ] - KL(q(z,c|x)||p(z,c))
              = Eq(z,c|x)[ log p(x|z) + log p(z|c) + log p(c) - log q(z|x) - log q(c|x) ]
    """
    mu_c, var_c, pi = c_params;  # print(mu_c.size(), var_c.size(), pi.size())
    n_centroids = pi.size(1)
    mu, logvar = z_params
    mu_expand = mu.unsqueeze(2).expand(mu.size(0), mu.size(1), n_centroids)
    logvar_expand = logvar.unsqueeze(2).expand(logvar.size(0), logvar.size(1), n_centroids)

    # log p(z|c)
    logpzc = -0.5 * torch.sum(gamma * torch.sum(math.log(2 * math.pi) + \
                                                torch.log(var_c) + \
                                                torch.exp(logvar_expand) / var_c + \
                                                (mu_expand - mu_c) ** 2 / var_c, dim=1), dim=1)
    # log p(c)
    logpc = torch.sum(gamma * torch.log(pi), 1)

    # log q(z|x) or q entropy
    qentropy = -0.5 * torch.sum(1 + logvar + math.log(2 * math.pi), 1)

    # log q(c|x)
    logqcx = torch.sum(gamma * torch.log(gamma), 1)

    kld = -logpzc - logpc + qentropy + logqcx

    return kld

####################################
## Created by DCCA
####################################

class Eucli_dis(nn.Module):
    """like what you like: knowledge distill via neuron selectivity transfer"""

    def __init__(self):
        super(Eucli_dis, self).__init__()
        pass

    def forward(self, g_s, g_t):
        g_s = g_s.float()
        g_t = g_t.float()
        ret = torch.pow((g_s - g_t), 2)

        return torch.sum(ret, dim=1)


class L1_dis(nn.Module):
    """like what you like: knowledge distill via neuron selectivity transfer"""

    def __init__(self):
        super(L1_dis, self).__init__()
        pass

    def forward(self, g_s, g_t):
        g_s = g_s.float()
        g_t = g_t.float()

        ret = torch.abs(g_s - g_t)

        return torch.sum(ret, dim=1)


class NSTLoss(nn.Module):
    """like what you like: knowledge distill via neuron selectivity transfer"""

    def __init__(self):
        super(NSTLoss, self).__init__()
        pass

    def forward(self, g_s, g_t):
        return [self.nst_loss(f_s, f_t) for f_s, f_t in zip(g_s, g_t)]

    def nst_loss(self, f_s, f_t):

        f_s = f_s.view(f_s.shape[0], f_s.shape[1], -1)
        f_s = F.normalize(f_s, dim=2)
        f_t = f_t.view(f_t.shape[0], f_t.shape[1], -1)
        f_t = F.normalize(f_t, dim=2)

        # set full_loss as False to avoid unnecessary computation
        full_loss = False
        if full_loss:
            return (self.poly_kernel(f_t, f_t).mean().detach() + self.poly_kernel(f_s, f_s).mean()
                    - 2 * self.poly_kernel(f_s, f_t).mean())
        else:
            return self.poly_kernel(f_s, f_s).mean() - 2 * self.poly_kernel(f_s, f_t).mean()

    def poly_kernel(self, a, b):
        a = a.unsqueeze(1)
        b = b.unsqueeze(2)
        res = (a * b).sum(-1).pow(2)
        return res


class FactorTransfer(nn.Module):
    """Paraphrasing Complex Network: Network Compression via Factor Transfer, NeurIPS 2018"""

    def __init__(self, p1=2, p2=1):
        super(FactorTransfer, self).__init__()
        self.p1 = p1
        self.p2 = p2

    def forward(self, f_s, f_t):
        return self.factor_loss(f_s, f_t)

    def factor_loss(self, f_s, f_t):
        s_H, t_H = f_s.shape[2], f_t.shape[2]
        if s_H > t_H:
            f_s = F.adaptive_avg_pool2d(f_s, (t_H, t_H))
        elif s_H < t_H:
            f_t = F.adaptive_avg_pool2d(f_t, (s_H, s_H))
        else:
            pass
        if self.p2 == 1:
            return (self.factor(f_s) - self.factor(f_t)).abs().mean()
        else:
            return (self.factor(f_s) - self.factor(f_t)).pow(self.p2).mean()

    def factor(self, f):
        return F.normalize(f.pow(self.p1).mean(1).view(f.size(0), -1))


class Similarity(nn.Module):
    """Similarity-Preserving Knowledge Distillation, ICCV2019, verified by original author"""

    def __init__(self):
        super(Similarity, self).__init__()

    def forward(self, g_s, g_t):
        return [self.similarity_loss(f_s, f_t) for f_s, f_t in zip(g_s, g_t)]

    def similarity_loss(self, f_s, f_t):
        bsz = f_s.shape[0]
        f_s = f_s.view(bsz, -1)
        f_t = f_t.view(bsz, -1)

        G_s = torch.mm(f_s, torch.t(f_s))
        # G_s = G_s / G_s.norm(2)
        G_s = torch.nn.functional.normalize(G_s)
        G_t = torch.mm(f_t, torch.t(f_t))
        # G_t = G_t / G_t.norm(2)
        G_t = torch.nn.functional.normalize(G_t)

        G_diff = G_t - G_s
        loss = (G_diff * G_diff).view(-1, 1).sum(0) / (bsz * bsz)
        return loss


class Correlation(nn.Module):
    """Correlation Congruence for Knowledge Distillation, ICCV 2019.
    The authors nicely shared the code with me. I restructured their code to be
    compatible with my running framework. Credits go to the original author"""

    def __init__(self):
        super(Correlation, self).__init__()

    def forward(self, f_s, f_t):
        delta = torch.abs(f_s - f_t)
        loss = torch.mean((delta[:-1] * delta[1:]).sum(1))
        return loss


class KL_diver(nn.Module):
    def __init__(self):
        super(KL_diver, self).__init__()

    def forward(self, mean_1, logvar_1, mean_2, logvar_2):
        loss = kl(Normal(mean_1, logvar_1),
                  Normal(mean_2, logvar_2)).sum(dim=1)

        return loss


class Attention(nn.Module):
    """Paying More Attention to Attention: Improving the Performance of Convolutional Neural Networks
    via Attention Transfer
    code: https://github.com/szagoruyko/attention-transfer"""

    def __init__(self, p=2):
        super(Attention, self).__init__()
        self.p = p

    def forward(self, g_s, g_t):
        g_s_norm = F.normalize(g_s, p=2, dim=1)
        g_t_norm = F.normalize(g_t, p=2, dim=1)
        diff_g = g_s_norm - g_t_norm

        result = (diff_g.norm(p=2, dim=1, keepdim=True)).sum(dim=1)

        return result


class ZINBLoss(nn.Module):
    """ZINB loss class.
    """

    def __init__(self):
        super(ZINBLoss, self).__init__()

    def forward(self, x, mean, disp, pi, scale_factor, ridge_lambda=0.0):
        """Forward propagation.

        Parameters
        ----------
        x :
            input features.
        mean :
            data mean.
        disp :
            data dispersion.
        pi :
            data dropout probability.
        scale_factor : list
            scale factor of mean.
        ridge_lambda : float optional
            ridge parameter.

        Returns
        -------
        result : float
            ZINB loss.

        """
        eps = 1e-10
        scale_factor = scale_factor[:, None]
        mean = mean * scale_factor

        t1 = torch.lgamma(disp + eps) + torch.lgamma(x + 1.0) - torch.lgamma(x + disp + eps)
        t2 = (disp + x) * torch.log(1.0 + (mean / (disp + eps))) + (x * (torch.log(disp + eps) - torch.log(mean + eps)))
        nb_final = t1 + t2

        nb_case = nb_final - torch.log(1.0 - pi + eps)
        zero_nb = torch.pow(disp / (disp + mean + eps), disp)
        zero_case = -torch.log(pi + ((1.0 - pi) * zero_nb) + eps)
        result = torch.where(torch.le(x, 1e-8), zero_case, nb_case)

        if ridge_lambda > 0:
            ridge = ridge_lambda * torch.square(pi)
            result += ridge

        result = torch.mean(result)
        return result


def dist_loss(data, min_dist, max_dist=20):
    pairwise_dist = cdisttf(data, data)
    dist = pairwise_dist - min_dist
    bigdist = max_dist - pairwise_dist
    loss = torch.exp(-dist) + torch.exp(-bigdist)
    return loss

def cdisttf(data_1, data_2):
    prod = torch.sum((data_1.unsqueeze(1) - data_2.unsqueeze(0)) ** 2, 2)
    return (prod + 1e-10) ** (1 / 2)


import torch.nn as nn
import torch
import math
import torch.optim as optim
import numpy as np
import os
from torch.utils.data import DataLoader
from dance.utils.metrics import batch_separated_bipartite_matching
from dance.utils import SimpleIndexDataset

def propagation_layer_combination(X, Y, idx, wt1, wt2, from_logits=True):
    if from_logits:
        wt1 = torch.softmax(wt1, -1)
    x = 0
    for i in range(wt1.shape[0]):
        x += wt1[i] * X[i][idx]

    if from_logits:
        wt2 = torch.softmax(wt2, -1)
    y = 0
    for i in range(wt2.shape[0]):
        y += wt2[i] * Y[i][idx]
    return x, y

class ScMoGCNWrapper(object):
    """ScMoGCN class.

    Parameters
    ----------
    args : argparse.Namespace
        A Namespace object that contains arguments of ScMoGCN. For details of parameters in parser args, please refer to link (parser help document).
    layers : list[int]
        Specification of dimensions of hidden layers.
    temp : int optional
        Temperature for softmax, by default to be 1.
    """

    def __init__(self, args, layers, temp=1):
        self.model = ScMoGCN(args, layers, temp).to(args.device)
        self.args = args
        wt1 = torch.tensor([0.] * (args.layers - 1)).to(args.device).requires_grad_(True)
        wt2 = torch.tensor([0.] * (args.layers - 1)).to(args.device).requires_grad_(True)
        self.wt = [wt1, wt2]

    def to(self, device):
        """Performs device conversion.

        Parameters
        ----------
        device : str
            Target device.

        Returns
        -------
        self : ScMoGCNWrapper
            Converted model.
        """
        self.args.device = device
        self.model = self.model.to(device)

    def load(self, path, map_location=None):
        """Load model parameters from checkpoint file.

       Parameters
       ----------
       path : str
           Path to the checkpoint file.
       map_location : str optional
           Mapped device. This parameter will be passed to torch.load function if not none.

       Returns
       -------
       None.

       """
        if map_location is not None:
            self.model.load_state_dict(torch.load(path, map_location=map_location))
        else:
            self.model.load_state_dict(torch.load(path))

    def fit(self, dataset, feats, labels):
        """fit function for training.

        Parameters
        ----------
        dataset : dance.datasets.multimodality.ModalityMatchingNIPSDataset
            Dataset for mdality matching.
        feats : torch.Tensor
            Modality features.

        Returns
        -------
        None.
        """

        device = self.args.device
        wt = self.wt
        hcell_mod1, hcell_mod2 = feats

        labels0, labels1 = labels

        criterion = nn.CrossEntropyLoss()
        criterion2 = nn.MSELoss()

        assert wt[0].requires_grad == wt[1].requires_grad
        opt = optim.AdamW([{'params': self.model.parameters()}, {'params': wt[0]}, {'params': wt[1]}], lr=self.args.learning_rate)

        BATCH_SIZE = 4096

        idx = torch.randperm(dataset.sparse_features()[0].shape[0])
        train_idx = idx[:-BATCH_SIZE]
        val_idx = idx[-BATCH_SIZE:]
        test_idx = np.arange(dataset.sparse_features()[0].shape[0],
                             dataset.sparse_features()[0].shape[0] + dataset.sparse_features()[2].shape[0])
        train_dataset = SimpleIndexDataset(train_idx)
        train_loader = DataLoader(
            dataset=train_dataset,
            batch_size=BATCH_SIZE,
            shuffle=True,
            num_workers=0,
            drop_last=True,
        )

        maxval = -1
        vals = []
        for epoch in range(self.args.epochs):
            self.model.train()
            print('epoch', epoch)
            total_loss = 0
            accum_acc = [0, 0]

            for step, batch_idx in enumerate(train_loader):
                X, Y = propagation_layer_combination(hcell_mod1, hcell_mod2, batch_idx, wt[0], wt[1])

                logits = self.model(X, Y)
                temp = torch.arange(logits.shape[0]).to(logits.device)
                loss = criterion(logits, temp) + criterion(logits.T, temp)

                forward_accuracy = (torch.argmax(logits, dim=1) == temp).float().mean().item()
                backward_accuracy = (torch.argmax(logits, dim=0) == temp).float().mean().item()
                accum_acc[0] += forward_accuracy
                accum_acc[1] += backward_accuracy

                emb1, emb2 = self.model.encode(X, Y)
                pred1, pred2 = self.model.decode(emb2, emb1)
                rec1, rec2 = self.model.decode(emb1, emb2)

                loss2 = criterion2(pred1, X) + criterion2(pred2, Y)
                loss3 = criterion2(rec1, X) + criterion2(rec2, Y)

                total_loss += loss.item()

                if self.args.auxiliary_loss > 0:
                    loss = loss + loss2 + loss3

                opt.zero_grad()
                loss.backward()
                opt.step()

            print('training loss: %.5f, foward: %.4f, backward: %.4f' % (
                total_loss / len(train_loader), accum_acc[0] / len(train_loader), accum_acc[1] / len(train_loader)))

            temp = torch.arange(val_idx.shape[0]).to(device)
            vals.append(self.score([hcell_mod1, hcell_mod2], val_idx, [temp, temp]))
            print('validation score: %.5f' % vals[-1])
            if epoch % 10 == 9:
                print('testing score: %.5f' % self.score([hcell_mod1, hcell_mod2], test_idx, [labels0, labels1]))

            if vals[-1] > maxval:
                maxval = vals[-1]
                if not os.path.exists('models'):
                    os.mkdir('models')
                torch.save(self.model.state_dict(), f'models/model_{self.args.rnd_seed}.pth')
                weight_record = [wt[0].detach(), wt[1].detach()]

            if max(vals) != max(vals[-20:]):
                print('Early stopped.')
                break

        print('Valid: ', maxval)

        self.wt = weight_record
        return self

    def predict(self, inputs, idx, enhance = False, dataset = None):
        """Predict function to get latent representation of data.

        Parameters
        ----------
        inputs : list[torch.Tensor]
            Multimodality features.
        idx : Iterable[int]
            Cell indices for prediction.
        enhance : bool optional
            Whether enable enhancement matching (e.g. bipartite matching), by default to be False.
        dataset : dance.datasets.multimodality.ModalityMatchingNIPSDataset optional
            Dataset for modality matching, needed when enhance parameter set to be True.

        Returns
        -------
        pred : torch.Tensor
            Predicted matching matrix.

        """
        # inputs: [train_mod1, train_mod2], idx: valid_idx, labels: [sol, sol.T], wt: [wt0, wt1]
        self.model.eval()

        with torch.no_grad():
            wt = self.wt
            m1, m2 = propagation_layer_combination(inputs[0], inputs[1], idx, wt[0], wt[1])

            if not enhance:
                pred = self.model(m1, m2)
                return pred

            else:
                emb1, emb2 = self.model.encode(m1, m2)
                pred = batch_separated_bipartite_matching(dataset, emb1, emb2)
                return pred

    def score(self, inputs, idx, labels, enhance = False, dataset = None):
        """Score function to get score of prediction.

        Parameters
        ----------
        inputs : list[torch.Tensor]
            Multimodality features.
        idx : Iterable[int]
            Index of testing cells for scoring.
        labels : torch.Tensor
            Ground truth label of cell matching matrix
        enhance : bool optional
            Whether enable enhancement matching (e.g. bipartite matching), by default to be False.
        dataset : dance.datasets.multimodality.ModalityMatchingNIPSDataset optional
            Dataset for modality matching, needed when enhance parameter set to be True.

        Returns
        -------
        score : float
            Accuracy of predicted matching between two modalities.

        """

        if not enhance:

            logits = self.predict(inputs, idx, enhance, dataset)
            forward_accuracy = (torch.argmax(logits, dim=1) == labels[1]).float().mean().item()
            backward_accuracy = (torch.argmax(logits, dim=0) == labels[0]).float().mean().item()

            return (forward_accuracy + backward_accuracy) / 2

        else:

            matrix = self.predict(inputs, idx, enhance, dataset)
            score = (matrix * labels.numpy()).sum() / labels.shape[0]

            return score


class ScMoGCN(nn.Module):
    def __init__(self, args, layers, temp=1):
        super().__init__()

        assert(len(layers) == 4)
        self.nn = [list() for i in range(4)]
        self.temp = temp
        self.args = args

        for j, shape in enumerate(layers):
            for i, s in enumerate(shape):
                self.nn[j].append(nn.Linear(s[0], s[1]))
                if i < len(shape)-1:
                    self.nn[j].append(nn.GELU())
                    if len(s) == 3:
                        self.nn[j].append(nn.Dropout(s[2]))

        self.nn = [nn.Sequential(*n) for n in self.nn]
        for i, n in enumerate(self.nn):
            self.add_module(str(i), n)


    def encode(self, m1, m2):
        emb1 = self.nn[0](m1)
        emb2 = self.nn[1](m2)
        emb1 = emb1 / torch.norm(emb1, p=2, dim=-1, keepdim=True)
        emb2 = emb2 / torch.norm(emb2, p=2, dim=-1, keepdim=True)
        return emb1, emb2

    def decode(self, emb1, emb2):
        return self.nn[2](emb1), self.nn[3](emb2)

    def forward(self, m1, m2):
        emb1, emb2 = self.encode(m1, m2)
        return torch.matmul(emb1, emb2.T) * math.exp(self.temp)






"""
This is the implementation of STAGATE.
Extended from https://github.com/QIFEIDKN/STAGATE

Reference
----------
Dong, Kangning, and Shihua Zhang. "Deciphering spatial domains from spatially resolved transcriptomics with an adaptive graph attention auto-encoder." Nature communications 13.1 (2022): 1-12.
"""

import pandas as pd
import numpy as np
import sklearn.neighbors
import scipy.sparse as sp
# import seaborn as sns
import matplotlib.pyplot as plt

import torch
from torch_geometric.data import Data
from torch import Tensor
import torch.nn.functional as F
from torch.nn import Parameter
import torch.nn as nn
from torch_sparse import SparseTensor, set_diag
from torch_geometric.nn.dense.linear import Linear
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.utils import remove_self_loops, add_self_loops, softmax
from tqdm import tqdm
import scanpy as sc
from sklearn import mixture


def transfer_pytorch_data(adata, adj):
    edgeList = adj
    if type(adata.X) == np.ndarray:
        data = Data(edge_index=torch.LongTensor(np.array(
            [edgeList[0], edgeList[1]])), x=torch.FloatTensor(adata.X))  # .todense()
    else:
        data = Data(edge_index=torch.LongTensor(np.array(
            [edgeList[0], edgeList[1]])), x=torch.FloatTensor(adata.X.todense()))  # .todense()
    return data

def Stats_Spatial_Net(adata):
    import matplotlib.pyplot as plt
    Num_edge = adata.uns['Spatial_Net']['Cell1'].shape[0]
    Mean_edge = Num_edge/adata.shape[0]
    plot_df = pd.value_counts(pd.value_counts(adata.uns['Spatial_Net']['Cell1']))
    plot_df = plot_df/adata.shape[0]
    fig, ax = plt.subplots(figsize=[3,2])
    plt.ylabel('Percentage')
    plt.xlabel('')
    plt.title('Number of Neighbors (Mean=%.2f)'%Mean_edge)
    ax.bar(plot_df.index, plot_df)

def mclust_P(adata, num_cluster, used_obsm='STAGATE', modelNames='EEE'):
    
    from sklearn import mixture
    np.random.seed(2020)
    g = mixture.GaussianMixture(n_components=num_cluster, covariance_type='spherical')
    res = g.fit_predict(adata.obsm[used_obsm])
    adata.obs['mclust'] = res
    return adata

'''
def mclust_R(adata, num_cluster, modelNames='EEE', used_obsm='STAGATE', random_seed=2020):
    """\
    Clustering using the mclust algorithm.
    The parameters are the same as those in the R package mclust.
    """
    
    np.random.seed(random_seed)
    import rpy2.robjects as robjects
    robjects.r.library("mclust")

    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)
    rmclust = robjects.r['Mclust']

    res = rmclust(rpy2.robjects.numpy2ri.numpy2rpy(adata.obsm[used_obsm]), num_cluster, modelNames)
    mclust_res = np.array(res[-2])

    adata.obs['mclust'] = mclust_res
    adata.obs['mclust'] = adata.obs['mclust'].astype('int')
    adata.obs['mclust'] = adata.obs['mclust'].astype('category')
    return adata
'''

class GATConv(MessagePassing):
    """
    Graph attention layer from Graph Attention Network.
    """
    _alpha = None
    def __init__(self, in_channels,
                 out_channels, heads: int = 1, concat: bool = True,
                 negative_slope: float = 0.2, dropout: float = 0.0,
                 add_self_loops= True, bias = True, **kwargs):
        kwargs.setdefault('aggr', 'add')
        super(GATConv, self).__init__(node_dim=0, **kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.heads = heads
        self.concat = concat
        self.negative_slope = negative_slope
        self.dropout = dropout
        self.add_self_loops = add_self_loops


        self.lin_src = nn.Parameter(torch.zeros(size=(in_channels, out_channels)))
        nn.init.xavier_normal_(self.lin_src.data, gain=1.414)
        self.lin_dst = self.lin_src


        # The learnable parameters to compute attention coefficients:
        self.att_src = Parameter(torch.Tensor(1, heads, out_channels))
        self.att_dst = Parameter(torch.Tensor(1, heads, out_channels))
        nn.init.xavier_normal_(self.att_src.data, gain=1.414)
        nn.init.xavier_normal_(self.att_dst.data, gain=1.414)

        self._alpha = None
        self.attentions = None



    def forward(self, x, edge_index,
                size=None, return_attention_weights=None, attention=True, tied_attention = None):

        H, C = self.heads, self.out_channels

        # We first transform the input node features. If a tuple is passed, we
        # transform source and target node features via separate weights:
        if isinstance(x, Tensor):
            assert x.dim() == 2, "Static graphs not supported in 'GATConv'"
            # x_src = x_dst = self.lin_src(x).view(-1, H, C)
            x_src = x_dst = torch.mm(x, self.lin_src).view(-1, H, C)
        else:  # Tuple of source and target node features:
            x_src, x_dst = x
            assert x_src.dim() == 2, "Static graphs not supported in 'GATConv'"
            x_src = self.lin_src(x_src).view(-1, H, C)
            if x_dst is not None:
                x_dst = self.lin_dst(x_dst).view(-1, H, C)

        x = (x_src, x_dst)

        if not attention:
            return x[0].mean(dim=1)
            # return x[0].view(-1, self.heads * self.out_channels)

        if tied_attention == None:
            # Next, we compute node-level attention coefficients, both for source
            # and target nodes (if present):
            alpha_src = (x_src * self.att_src).sum(dim=-1)
            alpha_dst = None if x_dst is None else (x_dst * self.att_dst).sum(-1)
            alpha = (alpha_src, alpha_dst)
            self.attentions = alpha
        else:
            alpha = tied_attention


        if self.add_self_loops:
            if isinstance(edge_index, Tensor):
                # We only want to add self-loops for nodes that appear both as
                # source and target nodes:
                num_nodes = x_src.size(0)
                if x_dst is not None:
                    num_nodes = min(num_nodes, x_dst.size(0))
                num_nodes = min(size) if size is not None else num_nodes
                edge_index, _ = remove_self_loops(edge_index)
                edge_index, _ = add_self_loops(edge_index, num_nodes=num_nodes)
            elif isinstance(edge_index, SparseTensor):
                edge_index = set_diag(edge_index)

        out = self.propagate(edge_index, x=x, alpha=alpha, size=size)

        alpha = self._alpha
        assert alpha is not None
        self._alpha = None

        if self.concat:
            out = out.view(-1, self.heads * self.out_channels)
        else:
            out = out.mean(dim=1)

        if isinstance(return_attention_weights, bool):
            if isinstance(edge_index, Tensor):
                return out, (edge_index, alpha)
            elif isinstance(edge_index, SparseTensor):
                return out, edge_index.set_value(alpha, layout='coo')
        else:
            return out

    def message(self, x_j, alpha_j, alpha_i,
                index, ptr,
                size_i) :
        alpha = alpha_j if alpha_i is None else alpha_j + alpha_i

        # alpha = F.leaky_relu(alpha, self.negative_slope)
        alpha = torch.sigmoid(alpha)
        
        alpha = softmax(alpha, index, ptr, size_i)
        self._alpha = alpha  # Save for later use.
        alpha = F.dropout(alpha, p=self.dropout, training=self.training)
        return x_j * alpha.unsqueeze(-1)

    def __repr__(self):
        return '{}({}, {}, heads={})'.format(self.__class__.__name__,
                                             self.in_channels,
                                             self.out_channels, self.heads)

class Stagate(torch.nn.Module):
    """Stagate class.
    
    Parameters
    ----------
    hidden_dims : int
        hidden dimensions
        
    """
    def __init__(self, hidden_dims):
        super().__init__()

        [in_dim, num_hidden, out_dim] = hidden_dims
        self.conv1 = GATConv(in_dim, num_hidden, heads=1, concat=False,
                             dropout=0, add_self_loops=False, bias=False)
        self.conv2 = GATConv(num_hidden, out_dim, heads=1, concat=False,
                             dropout=0, add_self_loops=False, bias=False)
        self.conv3 = GATConv(out_dim, num_hidden, heads=1, concat=False,
                             dropout=0, add_self_loops=False, bias=False)
        self.conv4 = GATConv(num_hidden, in_dim, heads=1, concat=False,
                             dropout=0, add_self_loops=False, bias=False)

    def forward(self, features, edge_index):
        """forward function for training.
        
        Parameters
        ----------
        features :  
            node features.
        edge_index :  
            adjacent matrix.
       
        Returns
        -------
        h2 :  
            the second hidden layer.
        h4 :  
            the forth hidden layer.
        
        """
        h1 = F.elu(self.conv1(features, edge_index))
        h2 = self.conv2(h1, edge_index, attention=False)
        self.conv3.lin_src.data = self.conv2.lin_src.transpose(0, 1)
        self.conv3.lin_dst.data = self.conv2.lin_dst.transpose(0, 1)
        self.conv4.lin_src.data = self.conv1.lin_src.transpose(0, 1)
        self.conv4.lin_dst.data = self.conv1.lin_dst.transpose(0, 1)
        h3 = F.elu(self.conv3(h2, edge_index, attention=True,
                              tied_attention=self.conv1.attentions))
        h4 = self.conv4(h3, edge_index, attention=False)

        return h2, h4  # F.log_softmax(x, dim=-1)


    def fit(self, adata, graph, n_epochs=1, lr=0.001, key_added='STAGATE',
                gradient_clipping=5., pre_resolution=0.2, weight_decay=0.0001, verbose=True, 
                random_seed=0, save_loss=False, save_reconstrction=False, 
                device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')):
        """fit function for training.
        
        Parameters
        ----------
        adata :  
            input data.
        graph :  
            graph structure.
        n_epochs : int optional
            number of epochs.
        lr : float optional
            learning rate.
        key_added : str optional
            by default 'STAGATE'.
        gradient_clipping : float optional
            gradient clipping.
        pre_resolution : float optional
            pre resolution.
        weight_decay : float optional  
            weight decay.
        verbose : bool optional  
            verbose, by default to be True.
        random_seed : int optional  
            random seed by default to be 0.
        save_loss : bool optional 
            by default to be False.
        save_reconstrction : bool optional 
            by default to be False.
        device : str optional
            to indicate gpu or cpu device.
       
        Returns
        -------
        None.
        
        """
        adata.X = sp.csr_matrix(adata.X)
        
        if 'highly_variable' in adata.var.columns:
            adata_Vars =  adata[:, adata.var['highly_variable']]
        else:
            adata_Vars = adata

        if verbose:
            print('Size of Input: ', adata_Vars.shape)
        if 'Spatial_Net' not in adata.uns.keys():
            raise ValueError("Spatial_Net is not existed! Run Cal_Spatial_Net first!")

        data = transfer_pytorch_data(adata_Vars, graph)

        model = self.to(device)
        data = data.to(device)

        optimizer = torch.optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay)

        # loss_list = []
        for epoch in tqdm(range(1, n_epochs+1)):
            model.train()
            optimizer.zero_grad()
            z, out = model(data.x, data.edge_index)
            loss = F.mse_loss(data.x, out) 
            # loss_list.append(loss)
            loss.backward()
            print("train.... epoch {} loss {}".format(epoch, loss))
            torch.nn.utils.clip_grad_norm_(model.parameters(), gradient_clipping)
            optimizer.step()
        
        print("eval ....")
        model.eval()
        z, out = model(data.x, data.edge_index)
        
        STAGATE_rep = z.to('cpu').detach().numpy()
        adata.obsm[key_added] = STAGATE_rep

        if save_loss:
            adata.uns['STAGATE_loss'] = loss
        if save_reconstrction:
            ReX = out.to('cpu').detach().numpy()
            ReX[ReX<0] = 0
            adata.layers['STAGATE_ReX'] = ReX

        
        print("post process...")
        sc.pp.neighbors(adata, use_rep='STAGATE')
        sc.tl.umap(adata)
        #adata = mclust_R(adata, used_obsm='STAGATE', num_cluster=7)
        adata = mclust_P(adata, used_obsm='STAGATE', num_cluster=7)
        self.adata = adata
    
    def predict(self,):
        """prediction function.
        
        Parameters
        ----------
     
        Returns
        -------
        self.y_pred : 
            predicted label.
        
        """
        data_dropna = self.adata.obs.dropna()
        self.y_pred = data_dropna['mclust']
        self.target = data_dropna['ground_truth']
        return data_dropna['mclust']
    
    def score(self, y_true=None):
        """score function to get score of prediction.
        
        Parameters
        ----------
        y_true :  
            ground truth label.
            
        Returns
        -------
        score : float
            metric eval score.
        
        """
        from sklearn.metrics.cluster import adjusted_rand_score
        score = adjusted_rand_score(self.target, self.y_pred)
        print("ARI {}".format(adjusted_rand_score(self.target, self.y_pred)))
        return score



import numpy as np
import pandas as pd
import scanpy as sc
import anndata as ad
from sklearn.ensemble import RandomForestClassifier
import warnings
from dance.transforms.preprocess import *



class SingleCellNet():
    """
    Build the SingleCellNet model
    
    Parameters
    ----------
    cgenesA: np.array
        All classification genes
    xpairs: np.array
        All top gene pairs
    tspRF: RandomForestClassifier
        Initialized Random Forest Classifier
    """
    def __init__(self, cgenesA = None, xpairs = None, tspRF = None):
        self.cgenesA = cgenesA
        self.xpairs = xpairs
        self.tspRF = tspRF
        
    def randomize(self, expDat, num=50):
        """
        Randomization sampling
        
        Parameters
        ----------
        expDat: 
            Data to be shuffled
        num: int
            number of samples selected
        
        Returns
        ---------- 
        output: pd.DataFrame
            A DataFrame with "num" random selected of rows
            
        """
        temp=expDat.to_numpy()
        temp=np.array([np.random.choice(x, len(x), replace=False) for x in temp])
        temp=temp.T
        temp=np.array([np.random.choice(x, len(x), replace=False) for x in temp]).T
        return pd.DataFrame(data=temp, columns=expDat.columns).iloc[0:num,:]
    
    def sc_trans_rnaseq(self, aDat, total = 10000):
        """
        normalization to aDat
        
        Parameters
        ----------
        expDat: Union[AnnData, ndarray, spmatrix]
            Data to be shuffled
        total: float
            If None, after normalization, each cell has a total count equal to the median of the counts_per_cell before normalization.
            
        Return
        ----------
        aDat: AnnData
            transformer result
        """
        sc.pp.normalize_per_cell(aDat, counts_per_cell_after=total)
        sc.pp.log1p(aDat)
        sc.pp.scale(aDat, max_value=10)
        #return np.log(1+expCountDnW)
        return aDat
    
    def sc_makeClassifier(self, expTrain, genes, groups, nRand=70, ntrees=2000, stratify=False):
        """
        Build the random forest classifier
                
        Parameters
        ----------
        expTrain: pd.DataFrame
            input dataset
        genes: np.array
            genes selected to fit the random forest model
        groups: str
            cell type labels with respect to the input
        nRand: int
            the number of samples selected
        ntrees: int
            Number of decision trees in random forest
        stratify: bool
            whether we select balanced class weight in the random forest model
        
        Returns
        ---------- 
        output: A random forest classifier 
            
        """
        
        randDat = self.randomize(expTrain, num=nRand)
        expT = pd.concat([expTrain, randDat])
        allgenes = expT.columns.values
        missingGenes = np.setdiff1d(np.unique(genes), allgenes)
        ggenes= np.intersect1d(np.unique(genes), allgenes)
        if not stratify:
            clf = RandomForestClassifier(n_estimators=ntrees, random_state=100)
        else:
            clf = RandomForestClassifier(n_estimators=ntrees,class_weight="balanced", random_state=100)
        ggroups=np.append(np.array(groups), np.repeat("rand", nRand)).flatten()
        clf.fit(expT.loc[:,ggenes].to_numpy(), ggroups)
        return clf
    
    def fit(self, aTrain,dLevel,nTopGenes = 100,nTopGenePairs = 100,nRand = 100, nTrees = 1000,stratify=False,counts_per_cell_after=1e4, scaleMax=10, limitToHVG=False):
        """
        Train the SingleCellNet model(random forest)
        
        Parameters
        ----------
        aTrain: AnnData object
            Training set in AnnData object version
        dLevel: str
            The column name of the labels
        nTopGenes: int
            Top n genes selected during training
        nTopGenePairs: int
            Top n genes pairs selected during training
        nRand: int
            Samples for each training batch
        nTrees: int
            Number of decision trees
        stratify: bool
            whether we select balanced class weight in the random forest model
        counts_per_cell_after: float
            If None, after normalization, each cell has a total count equal to the median of the counts_per_cell before normalization.
        scaleMax: int
            Maximum scaling for data normalization
        limitToHVG: bool
            whether we limit the model on highly variable genes
            
        Returns
        ---------- 
        There will be no output, but the parameters in the model will be updated
        
        """
        warnings.filterwarnings('ignore')
        stTrain= aTrain.obs
        
        #expRaw = pd.DataFrame(data=aTrain.X.toarray(),  index= aTrain.obs.index.values, columns= aTrain.var.index.values)
        expRaw = pd.DataFrame(data=aTrain.X,  index= aTrain.obs.index.values, columns= aTrain.var.index.values)
        expRaw = expRaw.loc[stTrain.index.values]
    
        adNorm = aTrain.copy()
        sc.pp.normalize_per_cell(adNorm, counts_per_cell_after=counts_per_cell_after)
        sc.pp.log1p(adNorm)
    
        print("HVG")
        if limitToHVG:
            sc.pp.highly_variable_genes(adNorm, min_mean=0.0125, max_mean=4, min_disp=0.5)
            adNorm = adNorm[:, adNorm.var.highly_variable]
    
        sc.pp.scale(adNorm, max_value=scaleMax)
        expTnorm= pd.DataFrame(data=adNorm.X,  index= adNorm.obs.index.values, columns= adNorm.var.index.values)
        expTnorm=expTnorm.loc[stTrain.index.values]
    
        ### expTnorm= pd.DataFrame(data=aTrain.X,  index= aTrain.obs.index.values, columns= aTrain.var.index.values)
        ### expTnorm=expTnorm.loc[stTrain.index.values]
        print("Matrix normalized")
        ### cgenesA, grps, cgenes_list =findClassyGenes(expTnorm,stTrain, dLevel = dLevel, topX = nTopGenes)
        cgenesA, grps, cgenes_list =findClassyGenes(expTnorm,stTrain, dLevel = dLevel, topX = nTopGenes)
        print("There are ", len(cgenesA), " classification genes\n")
        ### xpairs= ptGetTop(expTnorm.loc[:,cgenesA], grps, cgenes_list, topX=nTopGenePairs, sliceSize=5000)
        xpairs= ptGetTop(expTnorm.loc[:,cgenesA], grps, cgenes_list, topX=nTopGenePairs, sliceSize=5000)
    
        print("There are", len(xpairs), "top gene pairs\n")
        pdTrain= query_transform(expRaw.loc[:,cgenesA], xpairs)
        print("Finished pair transforming the data\n")
        tspRF=self.sc_makeClassifier(pdTrain.loc[:, xpairs], genes=xpairs, groups=grps, nRand = nRand, ntrees = nTrees, stratify=stratify)
        
        self.cgenesA = cgenesA
        self.xpairs = xpairs
        self.tspRF = tspRF
    
    def predict(self, adata):
        """
        Do prediction with singlecellnet model
        
        Parameters
        ----------
        adata: Union[AnnData, ndarray, spmatrix]
            test data
                
        Returns
        ---------- 
            output: a certain adata class from scanpy with an extra column names SCN_class as prediction
        
        """
        
        cgenes = self.cgenesA
        xpairs = self.xpairs
        rf_tsp = self.tspRF
        classRes = self.scn_predict(cgenes, xpairs, rf_tsp, adata, nrand = 0)
        categories = classRes.columns.values
        adNew = ad.AnnData(classRes, obs=adata.obs, var=pd.DataFrame(index=categories))
        # adNew.obs['category'] =  classRes.idxmax(axis=1)
        adNew.obs['SCN_class'] =  classRes.idxmax(axis=1)
        return adNew
    
    def score_exp(self, adNew, dtype):
        """
        singlecellnet model evaluation
        
        Parameters
        ----------
        adNew: a certain adata class from scanpy
        dtype: column name in of the label groundtruth
        
        Returns
        ---------- 
        correct: int
            total number of correct prediction 
        acc: float
            prediction accuracy
            
        """
        correct = sum(np.array(adNew.obs[dtype])== np.array(adNew.obs['SCN_class']))
        acc = correct/len(adNew.obs)    
        return correct, acc
    
    def score(self, adNew, test_dataset,map,dtype):
        """
        singlecellnet model evaluation
        
        Parameters
        ----------
        adNew: AnnData
            a certain adata class from scanpy
        test_dataset: AnnData
            test data
        map: dictionary
            dictionary for label conversion
        dtype: string
            column name in of the label groundtruth
        
        Returns
        ---------- 
        correct: int
            total number of correct prediction 
        acc: float
            prediction accuracy
            
        """
        correct = 0
        for i, cell in enumerate(np.array(adNew.obs[dtype])):
            if np.array(adNew.obs.SCN_class)[i] in map[cell]:
                correct += 1
      
        return (correct/len(np.array(adNew.obs.SCN_class)))
        

            
    def add_classRes(self, adata: AnnData, adClassRes, copy=False) -> AnnData:
        """
        add adClassRes to adata.obs
        
        Parameters
        ----------
        adata: AnnData object
            Original Anndata
        adClassRes: AnnData object
            Adding another dataset into adata
            
        Returns
        ---------- 
            correct: total number of correct prediction 
            acc: prediction accuracy
            
        """
        cNames = adClassRes.var_names
        for cname in cNames:
            adata.obs[cname] = adClassRes[:,cname].X
        # adata.obs['category'] = adClassRes.obs['category']
        adata.obs['SCN_class'] = adClassRes.obs['SCN_class']
        return adata if copy else None
    
    def check_adX(self, adata: AnnData) -> AnnData:
        """
        convert the feature matrix within adata class to sparse csr matrix
        
        Parameters
        ----------
        adata: Anndata
            a certain adata class from scanpy
            
        Returns
        ---------- 
        No return 
        """
        from scipy import sparse
        if( isinstance(adata.X, np.ndarray)):
            adata.X = sparse.csr_matrix(adata.X)
    
    
    def scn_predict(self, cgenes, xpairs, rf_tsp, aDat, nrand = 2):
        """
        Prediction with random forest
        
        Parameters
        ----------
        cgenes: np.array
            Classification genes in the model
        xpairs: np.array
            Top gene pairs in the model
        rf_tsp: RandomForestClassifier
            Initialized Random Forest Classifier
        aDat: AnnData
            input feature
        nrand: int
            batch size during training
            
        Return
        ----------
        classRes_val: pd.DataFrame
            Prediction result
        """
        if isinstance(aDat.X,np.ndarray):
            # in the case of aDat.X is a numpy array 
            aDat.X = ad._core.views.ArrayView(aDat.X)
    ###    expDat= pd.DataFrame(data=aDat.X, index= aDat.obs.index.values, columns= aDat.var.index.values)
        expDat= pd.DataFrame(data=aDat.X.toarray(), index= aDat.obs.index.values, columns= aDat.var.index.values)
        expValTrans=query_transform(expDat.reindex(labels=cgenes, axis='columns', fill_value=0), xpairs)
        classRes_val=self.rf_classPredict(rf_tsp, expValTrans, numRand=nrand)
        return classRes_val
    
    def rf_classPredict(self, rfObj,expQuery,numRand=50):
        """
        Prediction with random forest
        
        Parameters
        ----------
        expQuery: np.array
            input features
        numRand: int
            batch size during training
            
        Return
        ----------
        xpreds: pd.DataFrame
            Prediction result
        """
        if numRand > 0 :
            randDat=self.randomize(expQuery, num=numRand)
            expQuery=pd.concat([expQuery, randDat])
        xpreds= pd.DataFrame(rfObj.predict_proba(expQuery), columns= rfObj.classes_, index=expQuery.index)
        return xpreds